"""This is IOS XR Classic driver implementation."""

from functools import partial
import re
import logging
from condoor.drivers.generic import Driver as Generic
from condoor import pattern_manager, TIMEOUT, EOF, ConnectionAuthenticationError, ConnectionError
from condoor.fsm import FSM
from condoor.actions import a_reload_na, a_send, a_send_boot, a_reconnect, a_send_username, a_send_password,\
    a_message_callback, a_return_and_reconnect, a_not_committed

logger = logging.getLogger(__name__)


class Driver(Generic):
    """This is a Driver class implementation for IOS XR Classic."""

    platform = 'XR'
    inventory_cmd = 'admin show inventory chassis'
    users_cmd = 'show users'
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'rommon', 'xml', 'qnx']
    prepare_terminal_session = ['terminal exec prompt no-timestamp', 'terminal len 0', 'terminal width 0']
    reload_cmd = 'admin reload location all'
    families = {
        "ASR9K": "ASR9K",
        "ASR-9": "ASR9K",
        "CRS": "CRS",
    }

    def __init__(self, device):
        """Initialize the IOS XR Classic driver object."""
        super(Driver, self).__init__(device)

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['XR', 'QNX'])

    def reload(self, reload_timeout, save_config):
        """Reload the device."""
        PROCEED = re.compile(re.escape("Proceed with reload? [confirm]"))
        DONE = re.compile(re.escape("[Done]"))
        CONFIGURATION_COMPLETED = re.compile("SYSTEM CONFIGURATION COMPLETED")
        CONFIGURATION_IN_PROCESS = re.compile("SYSTEM CONFIGURATION IN PROCESS")

        # CONSOLE = re.compile("ios con[0|1]/RS?P[0-1]/CPU0 is now available")
        CONSOLE = re.compile("ios con[0|1]/(?:RS?P)?[0-1]/CPU0 is now available")
        RECONFIGURE_USERNAME_PROMPT = "[Nn][Oo] root-system username is configured"
        ROOT_USERNAME_PROMPT = "Enter root-system username\: "
        ROOT_PASSWORD_PROMPT = "Enter secret( again)?\: "

        # BOOT=disk0:asr9k-os-mbi-6.1.1/0x100305/mbiasr9k-rsp3.vm,1; \
        # disk0:asr9k-os-mbi-5.3.4/0x100305/mbiasr9k-rsp3.vm,2;
        # Candidate Boot Image num 0 is disk0:asr9k-os-mbi-6.1.1/0x100305/mbiasr9k-rsp3.vm
        # Candidate Boot Image num 1 is disk0:asr9k-os-mbi-5.3.4/0x100305/mbiasr9k-rsp3.vm
        CANDIDATE_BOOT_IMAGE = "Candidate Boot Image num 0 is .*vm"
        NOT_COMMITTED = re.compile(re.escape("Some active software packages are not yet committed. Proceed?[confirm]"))
        RELOAD_NA = re.compile("Reload to the ROM monitor disallowed from a telnet line")
        #           0          1      2                3                   4                  5
        events = [RELOAD_NA, DONE, PROCEED, CONFIGURATION_IN_PROCESS, self.rommon_re, self.press_return_re,
                  #   6               7                   8                           9
                  CONSOLE, CONFIGURATION_COMPLETED, RECONFIGURE_USERNAME_PROMPT, ROOT_USERNAME_PROMPT,
                  #    10                    11              12     13          14           15
                  ROOT_PASSWORD_PROMPT, self.username_re, TIMEOUT, EOF, self.reload_cmd, CANDIDATE_BOOT_IMAGE,
                  #   16
                  NOT_COMMITTED]

        transitions = [
            (RELOAD_NA, [0], -1, a_reload_na, 0),
            (NOT_COMMITTED, [0], -1, a_not_committed, 0),
            (DONE, [0], 2, None, 120),
            (PROCEED, [2], 3, partial(a_send, "\r"), reload_timeout),
            # this needs to be verified
            (self.rommon_re, [0, 3], 3, partial(a_send_boot, "boot"), 600),
            (CANDIDATE_BOOT_IMAGE, [0, 3], 4, a_message_callback, 600),
            (CONSOLE, [0, 1, 3, 4], 5, None, 600),
            (self.press_return_re, [5], 6, partial(a_send, "\r"), 300),
            # configure root username and password the same as used for device connection.
            (RECONFIGURE_USERNAME_PROMPT, [6, 7, 10], 8, None, 10),
            (ROOT_USERNAME_PROMPT, [8], 9, partial(a_send_username, self.device.node_info.username), 1),
            (ROOT_PASSWORD_PROMPT, [9], 9, partial(a_send_password, self.device.node_info.password), 1),
            (CONFIGURATION_IN_PROCESS, [6, 9], 10, None, 180),
            (CONFIGURATION_COMPLETED, [10], -1, a_reconnect, 0),
            (self.username_re, [7, 9], -1, a_return_and_reconnect, 0),
            (TIMEOUT, [0, 1, 2], -1, ConnectionAuthenticationError("Unable to reload"), 0),
            (EOF, [0, 1, 2, 3, 4, 5], -1, ConnectionError("Device disconnected"), 0),
            (TIMEOUT, [6], 7, partial(a_send, "\r"), 180),
            (TIMEOUT, [7, 10], -1, ConnectionAuthenticationError("Unable to reconnect after reloading"), 0),
        ]

        fsm = FSM("RELOAD", self.device, events, transitions, timeout=600)
        return fsm.run()



"""Provide the exception classes."""


class GeneralError(Exception):
    """General error."""

    def __init__(self, message=None, host=None):
        """The class constructor.

        Args:
            message (str): Custom message to be passed to the exceptions. Defaults to *None*.
                If *None* then the general class *__doc__* is used.
            host (str): Custom string which can be used to enhance the exception message by adding the "`host`: "
                prefix to the message string. Defaults to *None*. If `host` is *None* then message stays unchanged.
        """
        self.message = message
        self.hostname = str(host) if host else None

    def __str__(self):
        """Return string representing the exception."""
        message = self.message or self.__class__.__doc__
        return "{}: {}".format(self.hostname, message) if self.hostname else message


class InvalidHopInfoError(GeneralError):
    """Invalid device connection parameters."""

    pass


class ConnectionError(GeneralError):
    """General connection error."""

    pass


class ConnectionAuthenticationError(ConnectionError):
    """Connection authentication error."""

    pass


class ConnectionTimeoutError(ConnectionError):
    """Connection timeout error."""

    pass


class CommandError(GeneralError):
    """Command execution error."""

    def __init__(self, message=None, host=None, command=None):
        """The class constructor.

        Args:
            message (str): Custom message to be passed to the exceptions. Defaults to *None*.
                If *None* then the general class *__doc__* is used.
            host (str): Custom string which can be used to enhance the exception message by adding the "`host`: "
                prefix to the message string. Defaults to *None*. If `host` is *None* then message stays unchanged.
            command (str): Custom string which can be used enhance the exception message by adding the
                "`command`" suffix to the message string. Defaults to *None*. If `command` is *None* then the message
                stays unchanged.
        """
        GeneralError.__init__(self, message, host)
        self.command = command

    def __str__(self):
        """Return string representative of the exception."""
        message = self.message or self.__class__.__doc__
        message = "{}: '{}'".format(message, self.command) \
            if self.command else message
        message = "{}: {}".format(self.hostname, message) \
            if self.hostname else message
        return message


class CommandSyntaxError(CommandError):
    """Command syntax error."""

    pass


class CommandTimeoutError(CommandError):
    """Command timeout error."""

    pass



"""Provides Console Conserver driver class."""

from functools import partial
import re
import logging
import pexpect

from condoor.fsm import FSM
from condoor.utils import pattern_to_str
from condoor.protocols.base import Protocol
from condoor.actions import a_send, a_send_password, a_authentication_error, a_unable_to_connect,\
    a_save_last_pattern, a_send_username

from condoor.exceptions import ConnectionError, ConnectionTimeoutError
from condoor.config import CONF

logger = logging.getLogger(__name__)


# Telnet connection initiated
ESCAPE_CHAR = re.compile(re.escape("[Enter `^Ec?' for help]"))
# Connection refused i.e. line busy on TS
CONNECTION_REFUSED = re.compile("Connection refused")
PASSWORD_OK = "[Pp]assword [Oo][Kk]"

_C = CONF['protocol']['console']


class Console(Protocol):
    """Telnet protocol implementation."""

    def __init__(self, device):
        """Initialize the Telnet object."""
        super(Console, self).__init__(device)

    def get_command(self):
        """Return the Console protocol specific command to connect."""
        return "console {}".format(self.hostname)

    def connect(self, driver):
        """Connect using console specific FSM."""
        #              0            1                    2                      3
        events = [ESCAPE_CHAR, driver.press_return_re, driver.standby_re, driver.username_re,
                  #            4                   5            6                     7
                  driver.password_re, driver.more_re, self.device.prompt_re, driver.rommon_re,
                  #       8                           9              10             11
                  driver.unable_to_connect_re, driver.timeout_re, pexpect.TIMEOUT, PASSWORD_OK]

        transitions = [
            (ESCAPE_CHAR, [0], 1, partial(a_send, "\r\n"), _C['esc_char_timeout']),
            (driver.press_return_re, [0, 1], 1, partial(a_send, "\r\n"), 10),
            (PASSWORD_OK, [0, 1], 1, partial(a_send, "\r\n"), 10),
            (driver.standby_re, [0, 5], -1, ConnectionError("Standby console", self.hostname), 0),
            (driver.username_re, [0, 1, 5, 6], -1, partial(a_save_last_pattern, self), 0),
            (driver.password_re, [0, 1, 5], -1, partial(a_save_last_pattern, self), 0),
            (driver.more_re, [0, 5], 7, partial(a_send, "q"), 10),
            # router sends it again to delete
            (driver.more_re, [7], 8, None, 10),
            # (prompt, [0, 1, 5], 6, partial(a_send, "\r\n"), 10),
            (self.device.prompt_re, [0, 5], 0, None, 10),
            (self.device.prompt_re, [1, 6, 8, 5], -1, partial(a_save_last_pattern, self), 0),
            (driver.rommon_re, [0, 1, 5], -1, partial(a_save_last_pattern, self), 0),
            (driver.unable_to_connect_re, [0, 1], -1, a_unable_to_connect, 0),
            (driver.timeout_re, [0, 1], -1, ConnectionTimeoutError("Connection Timeout", self.hostname), 0),
            (pexpect.TIMEOUT, [0, 1], 5, partial(a_send, "\r\n"), 10),
            (pexpect.TIMEOUT, [5], -1, ConnectionTimeoutError("Connection timeout", self.hostname), 0)
        ]
        logger.debug("EXPECTED_PROMPT={}".format(pattern_to_str(self.device.prompt_re)))
        fsm = FSM("CONSOLE-SERVER-CONNECT", self.device, events, transitions, timeout=_C['connect_timeout'],
                  init_pattern=self.last_pattern)
        return fsm.run()

    def authenticate(self, driver):
        """Authenticate using the Console Server protocol specific FSM."""
        #                      0                      1                    2                    3
        events = [driver.username_re, driver.password_re, self.device.prompt_re, driver.rommon_re,
                  #       4             5                   6                       7                8
                  driver.unable_to_connect_re, driver.authentication_error_re, pexpect.TIMEOUT, pexpect.EOF]

        transitions = [
            (driver.username_re, [0], 1, partial(a_send_username, self.username), 10),
            (driver.username_re, [1], 1, None, 10),
            (driver.password_re, [0, 1], 2, partial(a_send_password, self._acquire_password()),
             _C['first_prompt_timeout']),
            (driver.username_re, [2], -1, a_authentication_error, 0),
            (driver.password_re, [2], -1, a_authentication_error, 0),
            (driver.authentication_error_re, [1, 2], -1, a_authentication_error, 0),
            (self.device.prompt_re, [0, 1, 2], -1, None, 0),
            (driver.rommon_re, [0], -1, partial(a_send, "\r\n"), 0),
            (pexpect.TIMEOUT, [0], 1, partial(a_send, "\r\n"), 10),
            (pexpect.TIMEOUT, [2], -1, None, 0),
            (pexpect.TIMEOUT, [3, 7], -1, ConnectionTimeoutError("Connection Timeout", self.hostname), 0),
            (driver.unable_to_connect_re, [0, 1, 2], -1, a_unable_to_connect, 0),
        ]
        logger.debug("EXPECTED_PROMPT={}".format(pattern_to_str(self.device.prompt_re)))
        fsm = FSM("CONSOLE-SERVER-AUTH", self.device, events, transitions, timeout=_C['connect_timeout'],
                  init_pattern=self.last_pattern)
        return fsm.run()

    def disconnect(self, device):
        """Disconnect using protocol specific method."""
        self.device.ctrl.sendcontrol('d')
        self.device.ctrl.send("c.")
        logger.debug("CONSOLE SERVER disconnect")
        try:
            self.device.ctrl.send(chr(4))
        except OSError:
            logger.debug("Protocol already disconnected")



"""Provides HopInfo class implementation and factory function."""

from urlparse import urlparse, parse_qs
from urllib import unquote

from condoor.exceptions import InvalidHopInfoError

# Standard protocol to port mapping
protocol2port_map = {
    'telnet': 23,
    'ssh': 22,
    'console': None,
}


def make_hop_info_from_url(url, verify_reachability=None):
    """Factory function to build HopInfo object from url.

    It allows only telnet and ssh as a valid protocols.

    Args:
        url (str): The url string describing the node. i.e.
            telnet://username@1.1.1.1. The protocol, username and address
            portion of url is mandatory. Port and password is optional.
            If port is missing the standard protocol -> port mapping is done.
            The password is optional i.e. for TS access directly to console
            ports.
            The path part is treated as additional password required for some
            systems, i.e. enable password for IOS devices.:
            telnet://<username>:<password>@<host>:<port>/<enable_password>
            <enable_password> is optional

        verify_reachability: This is optional callable returning boolean
            if node is reachable. It can be used to verify reachability
            of the node before making a connection. It can speedup the
            connection process when node not reachable especially with
            telnet having long timeout.

    Returns:
        HopInfo object or None if url is invalid or protocol not supported

    """
    parsed = urlparse(url)
    username = None if parsed.username is None else unquote(parsed.username)  # It's None if not exists
    password = None if parsed.password is None else unquote(parsed.password)  # It's None if not exists

    try:
        enable_password = parse_qs(parsed.query)["enable_password"][0]
    except KeyError:
        enable_password = None

    hop_info = HopInfo(
        parsed.scheme,
        parsed.hostname,
        username,
        password,
        parsed.port,
        enable_password,
        verify_reachability=verify_reachability
    )
    if hop_info.is_valid():
        return hop_info
    raise InvalidHopInfoError


class HopInfo(object):
    """HopInfo class contains implementation.

    It maintain all the information needed for node (jump host or device) access.
    """

    def __init__(
            self,
            protocol,
            hostname,
            username,
            password=None,
            port=None,
            enable_password=None,
            verify_reachability=None):
        """
        Initialize the HopInfo object.

        Args:
            protocol (str): 'telnet' or 'ssh'. The other protocols are not
                implemented.
            hostname (str): The hostname or IP address of the node
            username (str): The username for node access
            password (str): The password for provided username. This
                argument is optional and can be omitted. i.e. ssh hey auth
                or TS without authentication.
            port (number): Optional TCP port number. If not provided the
                the standard mapping for telnet and ssh is done automatically.
            verify_reachability (callable): This is optional callable returning
                True if if node is reachable. It can be used to verify
                reachability of the node before making a connection.
                It can speedup the connection process when node not
                reachable especially with telnet having long timeout.
        """
        self.hostname = hostname
        self.username = username
        self.password = password
        self.protocol = protocol
        self.enable_password = enable_password if enable_password != "" else None

        # if port not provided map port based on protocol standards
        self.port = port if port else \
            protocol2port_map.get(self.protocol, None)

        self.verify_reachability = verify_reachability

    def is_valid(self):
        """Return if protocol is valid."""
        if self.protocol not in ['telnet', 'ssh', 'console']:
            return False
        return True

    def is_reachable(self):
        """Return if host is reachable."""
        if self.verify_reachability and \
                hasattr(self.verify_reachability, '__call__'):
            return self.verify_reachability(host=self.hostname, port=self.port)
        # assume is reachable if can't verify
        return True

    def __repr__(self):
        """Return string representation of the class."""
        if self.username is None:
            repr_str = "{}://{}:{}".format(self.protocol, self.hostname, self.port)
        else:
            repr_str = "{}://{}@{}:{}".format(self.protocol, self.username, self.hostname, self.port)
        return repr_str



"""Protocols module providing different protocol handlers."""


from collections import defaultdict

from condoor.protocols.base import Protocol
from condoor.protocols.ssh import SSH
from condoor.protocols.telnet import Telnet
from condoor.protocols.telnet import TelnetConsole
from condoor.protocols.console import Console

protocol2object = defaultdict(
    Protocol, {
        'ssh': SSH,
        'telnet': Telnet,
        'telnet_console': TelnetConsole,
        'ssh_console': SSH,
        'console': Console,
        'console_console': Console,
    }
)


def make_protocol(protocol_name, device):
    """Factory function providing the Protocol object."""
    return protocol2object[protocol_name](device)



"""Provides Finite State Machine implementation."""

from inspect import isclass
from functools import wraps
import logging
from time import time

from pexpect import EOF
from condoor.exceptions import ConnectionError
from condoor.utils import pattern_to_str

logger = logging.getLogger(__name__)


def action(func):
    """Wrapper for FSM action function providing extended logging information based on doc string."""
    @wraps(func)
    def call_action(*args, **kwargs):
        """Wrap the function with logger debug."""
        if func.__doc__ is None:
            logger.debug("A={}".format(func.__name__))
        else:
            logger.debug("A={}".format(func.__doc__.split('\n', 1)[0]))
        return func(*args, **kwargs)
    return call_action


class FSM(object):
    """This class represents Finite State Machine for the current device connection.

    Here is the example of usage::

        to be done


    The example action::

        def send_newline(ctx):
            ctx.ctrl.sendline()
            return True

        def error(ctx):
            ctx.message = "Filesystem error"
            return False

        def readonly(ctx):
            ctx.message = "Filesystem is readonly"
            return False

    The ctx object description refer to :class:`condoor.controllers.fsm.FSM`.

    If the action returns True then the FSM continues processing. If the action returns False then FSM stops
    and the error message passed back to the ctx object is posted to the log.


    The FSM state is the integer number. The FSM starts with initial ``state=0`` and finishes if the ``next_state``
    is set to -1.

    If action returns False then FSM returns False. FSM returns True if reaches the -1 state.

    """

    class Context(object):
        """FSM Context class."""

        _slots__ = ('fsm_name', 'ctrl', 'event', 'state', 'finished', 'msg', 'pattern', 'device')

        def __init__(self, fsm_name, device):
            """Initialize the FSM context object.

            Args:
                fsm_name (str): Name of the FSM. This is used for logging.
                device (object): The device object.
            """
            self.device = device
            self.ctrl = device.ctrl
            self.fsm_name = fsm_name
            self.event = None
            self.state = 0
            self.finished = False
            self.msg = ""
            self.pattern = None

        def __str__(self):
            """Return the string representing the FSM context."""
            return "FSM Context:E={},S={},FI={},M='{}'".format(
                self.event, self.state, self.finished, self.msg)

    def __init__(self, name, device, events, transitions, init_pattern=None, timeout=300, searchwindowsize=-1,
                 max_transitions=20):
        """Initialize FSM object.

        Args:
            name (str): Name of the state machine used for logging purposes. Can't be *None*
            ctrl (object): Controller class representing the connection to the device
            events (list): List of expected strings or pexpect.TIMEOUT exception expected from the device.
            transitions (list): List of tuples in defining the state machine transitions.
            init_pattern (str): The pattern that was expected in the previous operation.
            timeout (int): Timeout between states in seconds. Defaults to 300 seconds.
            searchwindowsize (int): The size of search window. Defaults to -1.
            max_transitions (int): Max number of transitions allowed before quiting the FSM.

        The transition tuple format is as follows::

            (event, [list_of_states], next_state, action, timeout)

        - event (str): string from the `events` list which is expected to be received from device.
        - list_of_states (list): List of FSM states that triggers the action in case of event occurrence.
        - next_state (int): Next state for FSM transition.
        - action (func): function to be executed if the current FSM state belongs to `list_of_states` and the `event`
          occurred. The action can be also *None* then FSM transits to the next state without any action. Action
          can be also the exception, which is raised and FSM stops.
        """
        self.events = events
        self.device = device
        self.ctrl = device.ctrl
        self.timeout = timeout
        self.searchwindowsize = searchwindowsize
        self.name = name
        self.init_pattern = init_pattern
        self.max_transitions = max_transitions

        self.transition_table = self._compile(transitions, events)

    def _compile(self, transitions, events):
        compiled = {}
        for transition in transitions:
            event, states, new_state, act, timeout = transition
            if not isinstance(states, list):
                states = list(states)
            try:
                event_index = events.index(event)
            except ValueError:
                logger.debug("Transition for non-existing event: {}".format(
                    event if isinstance(event, str) else event.pattern))
            else:
                for state in states:
                    key = (event_index, state)
                    compiled[key] = (new_state, act, timeout)

        return compiled

    def run(self):
        """Start the FSM.

        Returns:
            boolean: True if FSM reaches the last state or false if the exception or error message was raised
        """
        ctx = FSM.Context(self.name, self.device)
        transition_counter = 0
        timeout = self.timeout
        logger.debug("{} Start".format(self.name))
        while transition_counter < self.max_transitions:
            transition_counter += 1
            try:
                start_time = time()
                if self.init_pattern is None:
                    ctx.event = self.ctrl.expect(self.events, searchwindowsize=self.searchwindowsize, timeout=timeout)
                else:
                    logger.debug("INIT_PATTERN={}".format(pattern_to_str(self.init_pattern)))
                    try:
                        ctx.event = self.events.index(self.init_pattern)
                    except ValueError:
                        logger.critical("INIT_PATTERN unknown.")
                        continue
                    finally:
                        self.init_pattern = None

                finish_time = time() - start_time
                key = (ctx.event, ctx.state)
                ctx.pattern = self.events[ctx.event]

                if key in self.transition_table:
                    transition = self.transition_table[key]
                    next_state, action_instance, next_timeout = transition
                    logger.debug("E={},S={},T={},RT={:.2f}".format(ctx.event, ctx.state, timeout, finish_time))
                    if callable(action_instance) and not isclass(action_instance):
                        if not action_instance(ctx):
                            logger.error("Error: {}".format(ctx.msg))
                            return False
                    elif isinstance(action_instance, Exception):
                        logger.debug("A=Exception {}".format(action_instance))
                        raise action_instance
                    elif action_instance is None:
                        logger.debug("A=None")
                    else:
                        logger.error("FSM Action is not callable: {}".format(str(action_instance)))
                        raise RuntimeWarning("FSM Action is not callable")

                    if next_timeout != 0:  # no change if set to 0
                        timeout = next_timeout
                    ctx.state = next_state
                    logger.debug("NS={},NT={}".format(next_state, timeout))

                else:
                    logger.warning("Unknown transition: EVENT={},STATE={}".format(ctx.event, ctx.state))
                    continue

            except EOF:
                raise ConnectionError("Session closed unexpectedly", self.ctrl.hostname)

            if ctx.finished or next_state == -1:
                logger.debug("{} Stop at E={},S={}".format(self.name, ctx.event, ctx.state))
                return True

        # check while else if even exists
        logger.error("FSM looped. Exiting")
        return False



"""This is IOS XE driver implementation."""

from functools import partial
import logging
import re

from condoor.drivers.IOS import Driver as IOSDriver
from condoor import pattern_manager, TIMEOUT, EOF
from condoor.actions import a_send_line, a_send, a_disconnect, a_message_callback, a_return_and_reconnect
from condoor.fsm import FSM

logger = logging.getLogger(__name__)


# based on IOS driver
class Driver(IOSDriver):
    """This is a Driver class implementation for IOS XR."""

    platform = 'XE'
    families = {
        "ASR-9": "ASR900",
    }

    def __init__(self, device):
        """Initialize the IOS XE driver object."""
        super(Driver, self).__init__(device)

    def update_driver(self, prompt):
        """Return driver name based on prompt analysis."""
        return pattern_manager.platform(prompt, ['XE'])

    def reload(self, reload_timeout=300, save_config=True):
        """Reload the device.

        CSM_DUT#reload

        System configuration has been modified. Save? [yes/no]: yes
        Building configuration...
        [OK]
        Proceed with reload? [confirm]
        """
        SAVE_CONFIG = re.compile(re.escape("System configuration has been modified. Save? [yes/no]: "))
        PROCEED = re.compile(re.escape("Proceed with reload? [confirm]"))
        IMAGE = re.compile("Passing control to the main image")
        BOOTSTRAP = re.compile("System Bootstrap")
        LOCATED = re.compile("Located .*")
        RETURN = re.compile(re.escape("Press RETURN to get started!"))

        response = "yes" if save_config else "no"

        #              0          1       2       3         4
        events = [SAVE_CONFIG, PROCEED, LOCATED, RETURN, self.username_re,
                  self.password_re, BOOTSTRAP, IMAGE, TIMEOUT, EOF]
        #              5              6          7       8      9

        transitions = [
            (SAVE_CONFIG, [0], 1, partial(a_send_line, response), 60),
            (PROCEED, [0, 1], 2, partial(a_send, "\r"), reload_timeout),
            (LOCATED, [2], 2, a_message_callback, reload_timeout),
            # if timeout try to send the reload command again
            (TIMEOUT, [0], 0, partial(a_send_line, self.reload_cmd), 10),
            (BOOTSTRAP, [2], -1, a_disconnect, reload_timeout),
            (IMAGE, [2], 3, a_message_callback, reload_timeout),
            (self.username_re, [3], -1, a_return_and_reconnect, 0),
            (self.password_re, [3], -1, a_return_and_reconnect, 0),
            (RETURN, [3], -1, a_return_and_reconnect, 0),
            (TIMEOUT, [2], -1, a_disconnect, 0),
            (EOF, [0, 1, 2, 3], -1, a_disconnect, 0)
        ]
        fsm = FSM("IOS-RELOAD", self.device, events, transitions, timeout=10)
        return fsm.run()



"""This provides the base Protocol class implementation."""

import logging

logger = logging.getLogger(__name__)


class Protocol(object):
    """Base Protocol class implementation."""

    def __init__(self, device):
        """Initialize the protocol object."""
        self.device = device
        self.port = self.device.node_info.port
        self.password = self.device.node_info.password
        self.username = self.device.node_info.username

        self.last_pattern = None
        self.matched_string = None

    def connect(self, driver):
        """Connect using specific protocol."""
        raise NotImplementedError("Connection method not implemented")

    def authenticate(self, driver):
        """Authenticate using specific protocol."""
        raise NotImplementedError("Authentication method not implemented")

    def disconnect(self, driver):
        """Disconnect using specific protocol."""
        raise NotImplementedError("Disconnect method not implemented")

    @property
    def hostname(self):
        """Return hostname from url."""
        return self.device.node_info.hostname

    def _acquire_password(self):
        # TODO: Remove
        password = self.password
        # if not password:
        #     if self.account_manager:
        #         self._dbg(20,
        #                   "{}: {}: Acquiring password for {} from system KeyRing".format(
        #                       self.protocol, self.hostname, self.username))
        #         password = self.account_manager.get_password(self.hostname, self.username, interact=True)
        #         if not password:
        #             self._dbg(30, "{}: {}: Password for {} does not exists in KeyRing".format(
        #                 self.protocol, self.hostname, self.username))
        return password



"""Provides a set of functions nad clases for different purpose."""

import logging
from logging.handlers import WatchedFileHandler
import socket
import codecs
import time
import re
import os
import yaml


def delegate(attribute_name, method_names):
    """Pass the call to the attribute called attribute_name for every method listed in method_names."""
    # hack for python 2.7 as nonlocal is not available
    info = {
        'attribute': attribute_name,
        'methods': method_names
    }

    def decorator(cls):
        """Decorate class."""
        attribute = info['attribute']
        if attribute.startswith("__"):
            attribute = "_" + cls.__name__ + attribute
        for name in info['methods']:
            setattr(cls, name, eval("lambda self, *a, **kw: "
                                    "self.{0}.{1}(*a, **kw)".format(attribute, name)))
        return cls
    return decorator


def to_list(item):
    """Convert to list.

    If the given item is iterable, this function returns the given item.
    If the item is not iterable, this function returns a list with only the
    item in it.

    @type  item: object
    @param item: Any object.
    @rtype:  list
    @return: A list with the item in it.
    """
    if hasattr(item, '__iter__'):
        return item
    return [item]


def is_reachable(host, port=23):
    """Check reachability for specified hostname/port.

    It tries to open TCP socket.
    It supports IPv6.
    :param host: hostname or ip address string
    :rtype: str
    :param port: tcp port number
    :rtype: number
    :return: True if host is reachable else false
    """
    try:
        addresses = socket.getaddrinfo(
            host, port, socket.AF_UNSPEC, socket.SOCK_STREAM
        )
    except socket.gaierror:
        return False

    for family, _, _, _, sockaddr in addresses:
        sock = socket.socket(family, socket.SOCK_STREAM)
        sock.settimeout(5)
        try:
            sock.connect(sockaddr)
        except IOError:
            continue

        sock.shutdown(socket.SHUT_RDWR)
        sock.close()
        # Wait 2 sec for socket to shutdown
        time.sleep(2)
        break
    else:
        return False
    return True


def pattern_to_str(pattern):
    """Convert regex pattern to string.

    If pattern is string it returns itself,
    if pattern is SRE_Pattern then return pattern attribute
    :param pattern: pattern object or string
    :return: str: pattern sttring
    """
    if isinstance(pattern, str):
        return repr(pattern)
    else:
        return repr(pattern.pattern) if pattern else None


def levenshtein_distance(str_a, str_b):
    """Calculate the Levenshtein distance between string a and b.

    :param str_a: String - input string a
    :param str_b: String - input string b
    :return: Number - Levenshtein Distance between string a and b
    """
    len_a, len_b = len(str_a), len(str_b)
    if len_a > len_b:
        str_a, str_b = str_b, str_a
        len_a, len_b = len_b, len_a
    current = range(len_a + 1)
    for i in range(1, len_b + 1):
        previous, current = current, [i] + [0] * len_a
        for j in range(1, len_a + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if str_a[j - 1] != str_b[i - 1]:
                change += + 1
            current[j] = min(add, delete, change)
    return current[len_a]


def parse_inventory(inventory_output=None):
    """Parse the inventory text and return udi dict."""
    udi = {
        "name": "",
        "description": "",
        "pid": "",
        "vid": "",
        "sn": ""
    }
    if inventory_output is None:
        return udi

    # find the record with chassis text in name or descr
    capture_next = False
    chassis_udi_text = None
    for line in inventory_output.split('\n'):
        lc_line = line.lower()
        if 'chassis' in lc_line and 'name' in lc_line and 'descr':
            capture_next = True
            chassis_udi_text = line
            continue
        if capture_next:
            inventory_output = chassis_udi_text + "\n" + line
            break

    match = re.search(r"(?i)NAME: (?P<name>.*?),? (?i)DESCR", inventory_output, re.MULTILINE)
    if match:
        udi['name'] = match.group('name').strip('" ,')

    match = re.search(r"(?i)DESCR: (?P<description>.*)", inventory_output, re.MULTILINE)
    if match:
        udi['description'] = match.group('description').strip('" ')

    match = re.search(r"(?i)PID: (?P<pid>.*?),? ", inventory_output, re.MULTILINE)
    if match:
        udi['pid'] = match.group('pid')

    match = re.search(r"(?i)VID: (?P<vid>.*?),? ", inventory_output, re.MULTILINE)
    if match:
        udi['vid'] = match.group('vid')

    match = re.search(r"(?i)SN: (?P<sn>.*)", inventory_output, re.MULTILINE)
    if match:
        udi['sn'] = match.group('sn')
    return udi


class FilteredFile(object):
    """Delegate class for handling filtered file object."""

    __slots__ = ['_file', '_pattern']

    def __init__(self, filename, mode="r", encoding=None, pattern=None):
        """Initialize FilteredFile object."""
        object.__setattr__(self, '_pattern', pattern)
        if encoding is None:
            object.__setattr__(self, '_file', open(filename, mode=mode))
        else:
            object.__setattr__(self, '_file', codecs.open(filename, mode=mode, encoding=encoding))

    def __getattr__(self, name):
        """Override standard getattr and delegate to file object."""
        return getattr(self._file, name)

    def __setattr__(self, name, value):
        """Override standard setattr and delegate to file object."""
        setattr(self._file, name, value)

    def write(self, text):
        """Override the standard write method to filter the content."""
        if self._pattern:
            # pattern already compiled no need to check
            result = re.search(self._pattern, text)
            if result:
                for group in result.groups():
                    if group:
                        text = text.replace(group, "***")
        self._file.write(text)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Custom context manager exit method."""
        self._file.close()

    def __enter__(self):
        """Custom context manager exit method."""
        return self


class FilteredFileHandler(WatchedFileHandler):
    """Class defining custom FileHandler for filtering sensitive information."""

    def __init__(self, filename, mode='a', encoding="utf-8", delay=0, pattern=None):
        """Initialize the FilteredFileHandler object."""
        self.pattern = pattern
        WatchedFileHandler.__init__(self, filename, mode=mode, encoding=encoding, delay=delay)

    def _open(self):
        return FilteredFile(self.baseFilename, mode=self.mode, encoding=self.encoding, pattern=self.pattern)


def normalize_urls(urls):
    """Overload urls and make list of lists of urls."""
    _urls = []
    if isinstance(urls, list):
        if urls:
            if isinstance(urls[0], list):
                # multiple connections (list of the lists)
                _urls = urls
            elif isinstance(urls[0], str):
                # single connections (make it list of the lists)
                _urls = [urls]
        else:
            raise RuntimeError("No target host url provided.")
    elif isinstance(urls, str):
        _urls = [[urls]]
    return _urls


def make_handler(log_dir, log_level):
    """Make logging handler."""
    if log_level > 0:
        if log_level == logging.DEBUG:
            formatter = logging.Formatter('%(asctime)-15s [%(levelname)8s] %(name)s:'
                                          '%(funcName)s(%(lineno)d): %(message)s')
        else:
            formatter = logging.Formatter('%(asctime)-15s [%(levelname)8s]: %(message)s')
        if log_dir:
            # Create the log directory.
            if not os.path.exists(log_dir):
                try:
                    os.makedirs(log_dir)
                except IOError:
                    log_dir = "./"
            log_filename = os.path.join(log_dir, 'condoor.log')
            # FIXME: take pattern from pattern manager
            handler = FilteredFileHandler(log_filename, pattern=re.compile("s?ftp://.*:(.*)@"))

        else:
            handler = logging.StreamHandler()

        handler.setFormatter(formatter)
    else:
        handler = logging.NullHandler()

    return handler


def yaml_file_to_dict(script_name, path=None):
    """Read yaml file and return the dict.

    It assumes the module file exists with the defaults.
    If the CONDOOR_{SCRIPT_NAME} env is set then the user file from the env is loaded and merged with the default

    There can be user file located in ~/.condoor directory with the {script_name}.yaml filename. If exists
    it is merget with default config.
    """
    def load_yaml(file_path):
        """Load YAML file from full file path and return dict."""
        with open(file_path, 'r') as yamlfile:
            try:
                dictionary = yaml.load(yamlfile)
            except yaml.YAMLError:
                return {}
        return dictionary

    def merge(user, default):
        """Merge two dicts."""
        if isinstance(user, dict) and isinstance(default, dict):
            for k, v in default.iteritems():
                if k not in user:
                    user[k] = v
                else:
                    user[k] = merge(user[k], v)
            return user
        else:
            return default

    if path is None:
        path = os.path.abspath('.')

    config_file_path = os.path.join(path, script_name + '.yaml')
    if not os.path.exists(config_file_path):
        raise RuntimeError('Config file does not exist: {}'.format(config_file_path))

    default_dict = load_yaml(config_file_path)

    user_config_file_path = os.path.join(os.path.expanduser('~'), '.condoor', script_name + '.yaml')
    user_config_file_path = os.getenv('CONDOOR_' + script_name.upper(), user_config_file_path)

    if os.path.exists(user_config_file_path):
        user_dict = load_yaml(user_config_file_path)
        default_dict = merge(user_dict, default_dict)

    return default_dict



"""This is jumphost driver class implementation."""

import re
import logging

from condoor.drivers.generic import Driver as Generic
from condoor import pattern_manager, CommandError

logger = logging.getLogger(__name__)


class Driver(Generic):
    """This is a Driver class implementation for Unix Jumphost."""

    platform = 'jumphost'
    inventory_cmd = None
    target_prompt_components = ['prompt_dynamic']
    prepare_terminal_session = []

    def __init__(self, device):
        """Initialize the Unix Jumphost driver object."""
        super(Driver, self).__init__(device)

    def get_version_text(self):
        """Return the version information from Unix host."""
        version_text = self.device.send('uname -sr', timeout=10)
        return version_text

    def update_hostname(self, prompt):
        """Return the hostname."""
        return self.device.hostname

    def get_hostname_text(self):
        """Return hostname information from the Unix host."""
        # FIXME: fix it, too complex logic
        try:
            hostname_text = self.device.send('hostname', timeout=10)
            if hostname_text:
                self.device.hostname = hostname_text.splitlines()[0]
                return hostname_text
        except CommandError:
            return None

    def make_dynamic_prompt(self, prompt):
        """Extend prompt with flexible mode handling regexp."""
        patterns = ["[\r\n]" + pattern_manager.pattern(
            self.platform, pattern_name, compiled=False) for pattern_name in self.target_prompt_components]

        patterns_re = "|".join(patterns).format(hostname=re.escape(prompt))

        try:
            prompt_re = re.compile(patterns_re)
        except re.error as e:  # pylint: disable=invalid-name
            raise RuntimeError("Pattern compile error: {} ({}:{})".format(e.message, self.platform, patterns_re))

        logger.debug("Dynamic prompt: '{}'".format(repr(prompt_re.pattern)))
        return prompt_re



"""Provides predefined actions for Finite State Machines."""
import logging
from condoor.fsm import action
from condoor.exceptions import ConnectionAuthenticationError, ConnectionError, ConnectionTimeoutError


@action
def a_send_line(text, ctx):
    """Send text line to the controller followed by `os.linesep`."""
    ctx.ctrl.sendline(text)
    return True


@action
def a_send(text, ctx):
    """Send text line to the controller."""
    ctx.ctrl.send(text)
    return True


@action
def a_send_username(username, ctx):
    """Sent the username text."""
    if username:
        ctx.ctrl.sendline(username)
        return True
    else:
        ctx.ctrl.disconnect()
        raise ConnectionAuthenticationError("Username not provided", ctx.ctrl.hostname)


@action
def a_send_password(password, ctx):
    """Send the password text.

    Before sending the password local echo is disabled.
    If password not provided it disconnects from the device and raises ConnectionAuthenticationError exception.
    """
    if password:
        ctx.ctrl.send_command(password, password=True)
        return True
    else:
        ctx.ctrl.disconnect()
        raise ConnectionAuthenticationError("Password not provided", ctx.ctrl.hostname)


@action
def a_authentication_error(ctx):
    """Raise ConnectionAuthenticationError exception and disconnect."""
    ctx.ctrl.disconnect()
    raise ConnectionAuthenticationError("Authentication failed", ctx.ctrl.hostname)


@action
def a_unable_to_connect(ctx):
    """Provide detailed information about the session (before, after) when unable to connect.

    The state machine finishes without exception
    """
    message = "{}{}".format(ctx.ctrl.before, ctx.ctrl.after)
    ctx.msg = message.strip().splitlines()[-1]
    ctx.device.last_error_msg = ctx.msg
    # ctx.msg = "{}{}".format(ctx.ctrl.before, ctx.ctrl.after)
    return False


@action
def a_standby_console(ctx):
    """Raise ConnectionError exception when connected to standby console."""
    ctx.device.is_console = True
    raise ConnectionError("Standby console", ctx.ctrl.hostname)


@action
def a_disconnect(ctx):
    """Disconnect from the device when device is reloading."""
    ctx.msg = "Device is reloading"
    ctx.ctrl.disconnect()
    return True


@action
def a_reload_na(ctx):
    """Provide the message when the reload is not possible."""
    ctx.msg = "Reload to the ROM monitor disallowed from a telnet line. " \
              "Set the configuration register boot bits to be non-zero."
    ctx.failed = True
    return False


@action
def a_not_committed(ctx):
    """Provide the message that current software is not committed and reload is not possible."""
    ctx.ctrl.sendline('n')
    ctx.msg = "Some active software packages are not yet committed. Reload may cause software rollback."
    ctx.device.chain.connection.emit_message(ctx.msg, log_level=logging.ERROR)
    ctx.failed = True
    return False


@action
def a_connection_closed(ctx):
    """Provide message when connection is closed by remote host."""
    ctx.msg = "Device disconnected"
    ctx.device.connected = False
    # do not stop FSM to detect the jumphost prompt
    return True


@action
def a_stays_connected(ctx):
    """Stay connected."""
    ctx.ctrl.connected = True
    ctx.device.connected = False
    return True


@action
def a_unexpected_prompt(ctx):
    """Provide message when received humphost prompt."""
    prompt = ctx.ctrl.after
    ctx.msg = "Received the jump host prompt: '{}'".format(prompt)
    ctx.device.connected = False
    ctx.finished = True
    raise ConnectionError("Unable to connect to the device.", ctx.ctrl.hostname)


@action
def a_connection_timeout(ctx):
    """Check the prompt and update the drivers."""
    prompt = ctx.ctrl.after
    ctx.msg = "Received the jump host prompt: '{}'".format(prompt)
    print(ctx.msg)
    ctx.device.connected = False
    ctx.finished = True
    raise ConnectionTimeoutError("Unable to connect to the device.", ctx.ctrl.hostname)


@action
def a_expected_prompt(ctx):
    """Update driver, config mode and hostname when received an expected prompt."""
    prompt = ctx.ctrl.after
    ctx.device.update_driver(prompt)
    ctx.device.update_config_mode()
    ctx.device.update_hostname()
    ctx.finished = True
    return True


@action
def a_save_last_pattern(obj, ctx):
    """Save last pattern in the context."""
    obj.last_pattern = ctx.pattern
    return True


@action
def a_send_boot(rommon_boot_command, ctx):
    """Send boot command."""
    ctx.ctrl.sendline(rommon_boot_command)
    return True


@action
def a_reconnect(ctx):
    """Reconnect."""
    ctx.device.connect(ctx.ctrl)
    return True


@action
def a_return_and_reconnect(ctx):
    """Send new line and reconnect."""
    ctx.ctrl.send("\r")
    ctx.device.connect(ctx.ctrl)
    return True


@action
def a_store_cmd_result(ctx):
    """Store the command result for complex state machines.

    It is useful when exact command output is embedded in another commands, i.e. admin show inventory in eXR.
    """
    result = ctx.ctrl.before
    # check if multi line
    index = result.find('\n')
    if index > 0:
        # remove first line
        result = result[index + 1:]
    ctx.device.last_command_result = result.replace('\r', '')
    return True


@action
def a_message_callback(ctx):
    """Message the captured pattern."""
    message = ctx.ctrl.after.strip().splitlines()[-1]
    ctx.device.chain.connection.emit_message(message, log_level=logging.INFO)
    return True



"""This is IOS XRv driver implementation."""

import logging
from condoor.drivers.XR import Driver as XR

logger = logging.getLogger(__name__)


class Driver(XR):
    """This is a Driver class implementation for IOS XRv."""

    platform = 'XRv'
    inventory_cmd = 'admin show inventory chassis'
    users_cmd = 'show users'
    target_prompt_components = ['prompt_dynamic', 'prompt_default', 'rommon', 'xml']
    prepare_terminal_session = ['terminal exec prompt no-timestamp', 'terminal len 0', 'terminal width 0']
    reload_cmd = 'admin reload location all'
    families = {
        "XRv": "IOS-XRv",
    }
